# DUMPB.AWK
# Andrew Schulman (andrew@ora.com), September 1996 
# requires LINK, DUMPBIN, DUMPPE
# 12/9/96 -- fixed a bunch of problems (thanks to Clive Turvey!)
# 12/16/96 -- added handling for ret_XXXX and jmp_XXXX labels
# 1/15/96 -- fixed problem only getting first set of imports; did cant_reach
# 1/18/96 -- fixes from Alex Shmidt for MSVC 4.2 link/dumpbin

# TODO: Function pointers!
# e.g., ref to 004013F4 before RegisterClassA, 
# push 4013F4 before CreateWindowExA,
# code @ 13F4 immediately follows a jmp, cant_reach

function hx(n)      { return sprintf("%0X", n); }

function exist(f)   { return (filesize(f) != -1); }

function fail(s)    { failed = 1; print s > stderr; exit; }

function msg(s)     { print s > stderr; }

function dbg(s)     { print s > stderr; }

function hex(s)     { return (0 + ("0x" s)); }

function strip1(p2) {
    if ($2 == p2) {
        getline;
        while ($2 == p2)
            getline;
        }
    }

function strip2(p2, p3) {
    if (($2 == p2) && ($3 == p3)) {
        getline;
        while (($2 == p2) && ($3 == p3))
            getline;
        }
    }

function len2(s)    { return (s ~ hex2); }

function no_ops() {
    strip1("CC");   # int 03
    strip2("00", "00");
    strip1("90"); # nop
    if (NF < 2)
        return;

    i = 2;
    while ((len2($i)) && ($i ~ hex2))
        i++;
    start = i;
    for (i=2, j=start; i<NF; i++, j++)
        $i = $j;
    NF--;
    }

function banner() {
    msg("DUMPB v. 0.65 Copyright (c) 1996 Andrew Schulman -- andrew@ora.com");
    }

BEGIN {
    banner();
    do_dump = 1;    ### TODO: make command-line switch

    if (ARGC < 2)
        fail("Win32 disassembly -- improves output from LINK /DUMPBIN /DISASM\n\
Requires LINK.EXE and DUMPPE.EXE\n\
Usage: dumpb [Win32 file] <sections> > output.asm");

    # hex = /^0[0-9A-F]+$/ ;
    hex = /^0[0-9A-F]+$|0[0-9A-F]+$/ ; # change by Alex Shmidt
    hex2 = /[0-9A-F][0-9A-F]/ ;
    hexc = /0[0-9a-fA-F]+:/ ;   # must have lower case!

    filename = ARGV[1];         # change so can have -dump switch!
    print "Run DUMPBIN /EXPORTS" > stderr;
    system("dumpbin /exports " filename " > __tmp.ex");
    if (! exist("__tmp.ex"))
        fail("DUMPBIN failed");
    system("echo DISASM START > __tmp.a");
    print "Run DUMPBIN /DISASM" > stderr;
    if (ARGC > 2) {
        for (i=2; i<ARGC; i++) {
            if (toupper(ARGV[i]) ~ /\-DUMP/)    # doesn't work!
                do_dump++;
            else {
                print "Section " ARGV[i] > stderr;
                system("dumpbin /section:" ARGV[i] " /disasm " filename " >>__tmp.a");
                }
            }
        }
    else 
        system("dumpbin /disasm " filename " >> __tmp.a");
    print "Run DUMPPE to get imports" > stderr;
    system("dumppe " filename " > __tmp.im");
    system("echo END CLIVE >> __tmp.im");
    if (! exist("__tmp.im"))
        fail("DUMPPE failed");
    if (ARGC == 2) {
        print filename;
        print "Run DUMPBIN to get data segment" > stderr;
        system("dumpbin /section:.data /rawdata:longs " filename " > __tmp.dat");
    }

    ARGV[1] = "__tmp.ex";
    ARGV[2] = "__tmp.im";
    ARGV[3] = "__tmp.a";    # two passes
    ARGV[4] = "__tmp.a";
    ARGC = 5;

    ARGV[5] = "__tmp.dat";
    ARGC ++;
    }

/Section contains the following Exports/ { in_exports = 1; }

(in_exports == 1) && ($1 ~ /ordinal/) { in_exports = 2; }

(in_exports == 2) && (NF == 4) {
    split($4, arr, "(");
    split(arr[2], arr, ")");
    export[hex(arr[1])] = $3;
    }

(in_exports == 2) && ($1 ~ /Summary/) { 
    in_exports = 0; 
    msg("Export table");
    }

(in_data == 0) && ($2 == "(hex)") && ($3 == "(dec)") { in_dumppe = 1; }

(in_dumppe == 1) && (/Image Base/) { image_base = hex($3); }

(in_dumppe == 1) && (/Imp Addr Hint Import/) { in_import = 1; }

(in_dumppe == 1) && (/Address of Entry Point/) { export[hex($5)] = "START"; }

(in_import == 1) && (NF == 3) && ($1 ~ /0[0-9A-F]*/) {  # Clive DUMPPE
    import[image_base + hex($1)] = $3;
    ## dbg("Added " $3);
    }

# (in_import == 1) && ((/IAT Entry/) || (NF == 0)) { # Clive -- NO!!!
#    in_dumppe = in_import = 0; }

(in_import == 1) && (/END CLIVE/) { in_dumppe = in_import = 0; }

# need better way to know when in disasm

(in_disasm == 0) && (/DISASM START/) { 
    in_disasm = 1; 
    getline;    # NB!!
    msg("Pass 1 -- Gather targets");
    }

(in_disasm == 1) && (/DISASM START/) { 
    in_disasm = 2; 
    msg("Pass 2 -- Generate output");
    }

(in_disasm == 2) && (/Dump of file/) { in_disasm = 3; }

(in_disasm == 1) && (NF==1) && ($1 ~ /\:$/) {
    split($1, arr, ":");
    symbol = arr[1];
    getline;
    addr = hex($1);
    export[addr] = symbol;
    }

(in_disasm == 1) && ($1 ~ hexc) {   # first pass - gather
    no_ops();

    if ($3 ~ hex) {
        addr = hex($3);
        if (($2 ~ /call/) && (! (addr in export)))
            export[addr] = "func_" hx(addr);
        else if (($2 ~ /j.*/) && (! (addr in loc)))
            loc[addr] = "loc_" hx(addr);
        }

    # NOT else! could be RET 04
    if ($2 ~ /ret/)
        ret[hex($1)]++;

    if ($2 ~ /jmp/) {
        if ($3 ~ hex)
            jmp[hex($1)] = hex($3);
        else if ($4 ~ /PTR/) {
            # if call DWORD PTR or j.* DWORD PTR
            # 00002d52:   jmp     DWORD PTR ds:[004060F4]
            split($5, arr, "[");
            split(arr[2], arr, "]");
            jmp[hex($1)] = hex(arr[1]);
            }
        }
}

function dump_tables() {
    print "\nIMAGE BASE"; print hx(image_base);
    print "\nEXPORT"; for (x in export) print hx(x), export[x];
    print "\nIMPORT"; for (x in import) print hx(x), import[x];
    print "\nLOC";    for (x in loc)    print hx(x), loc[x];
    print "\nJMP";    for (x in jmp)    print hx(x), jmp[x];
    print "\nRET";    for (x in ret)    print hx(x);
    print "\nMISSING";for (x in missing) print hx(x);
    print "\nCANT REACH"; for (x in notreach) print hx(x);
    }

function begin2() {
    print "Cleaning up jmp and ret labels" > stderr;

    for (x in jmp) {
        j = jmp[x];
        # may want to do this replacement recursively for jmp -> jmp 
        jmp[x] = "jmp_" ((j in import) ? import[j] :
                        (j in export) ? export[j] :
                        (j in loc)    ? loc[j] :
                        hx(j));

        # or have to do before check loc?
        if (j == x)                 jmp[x] = "HALT_" hx(x);

        # using possibly new jmp[x];
        if ((x in loc) && (loc[x] ~ /loc_/))        loc[x] = jmp[x];
        if ((x in export) && (export[x] ~ /func_/)) export[x] = jmp[x];
    }

    for (x in ret) {
        if ((x in loc) && (loc[x] ~ /loc_/))        loc[x] = "ret_" hx(x);
        if ((x in export) && (export[x] ~ /func_/)) export[x] = "ret_" hx(x);
        }

    # export, import labels generally better?
    for (x in loc) {
        if (x in export)        loc[x] = export[x];
        else if (x in import)   loc[x] = import[x];
        }

    did_begin2 = 1;
    }

(in_disasm == 3) && ($1 ~ hexc) {   # second pass - print
    if (! did_begin2)
        begin2();

    no_ops();

    split($1, arr, ":");
    addr = hex(arr[1]);
    addr2 = addr - image_base;
    if (addr in export) {
        print "";
        print export[addr] ":";
        }
    else if (addr2 in export)   { # Alex
        print "";
        print export[addr2] ":";
        }
    else if (addr in loc) {
        # no extra line is only difference for loc
        print loc[addr] ":";
        }
    else if (addr2 in loc) { # Alex??
        # no extra line is only difference for loc
        print loc[addr2] ":";
        }
    else if (cant_reach) {  # from prev line
        print "\n;; can't reach? function pointer?";
        # TODO: check this on pass 1, store in cant_reach table, look in
        # cant_reach table whenever mov DW or push??? Need to combine
        # with image base!
        notreach[addr]++;
        }

    if ($3 ~ hex) {
        addr = hex($3);
        if ($2 ~ /call/) {
            if (addr in export)
                sub($3, export[addr], $0);
            else
                missing[addr] ++;
            }
        else if ($2 ~ /j.*/) {
            if (addr in loc) { # this screws up on jmp to self! replace 2nd
                if (loc[addr] ~ /HALT_/)
                    $3 = loc[addr];
                else
                    sub($3, loc[addr], $0);
                }
            else
                missing[addr] ++;
            }
        }
    if ($4 ~ /PTR/i) { # do case-insensitive search: Clive
        $4 = "";
        if (match($3, /DWORD/i)) {  # could be mov eax,DWORD
            substr($3, RSTART, RLENGTH) = "DW";  # just replace matched part
            split($5, arr, "[");
            split(arr[2], arr, "]");
            addr = hex(arr[1]);
            if (addr in import)
                $5 = import[addr];
        }
        else if (match($3, /WORD/i)) 
            substr($3, RSTART, RLENGTH) = "W";
        else if (match($3, /BYTE/i)) 
            substr($3, RSTART, RLENGTH) = "B";
    }
    printf("\t%06lX\t%s\t", hex($1), $2);
    for (i=3; i<=NF; i++)
        printf("%s ", $i);
    printf("\n");

    cant_reach = (($2 ~ /ret/) || ($2 ~ /jmp/)); # for next line
}

($1 == "RAW" && $2 == "DATA") { in_data = 1; in_disasm = 0; }

(in_data == 1) { print; }

END {
    if (failed)     exit;

    if (do_dump)    dump_tables();

    for (i=1; i<=5; i++)
        if (ARGV[i] ~ /__tmp/)
            if (exist(ARGV[i]))
                system("del " ARGV[i]);
    }
